# Copyright 2010 Ben Escoto
#
# This file is part of FAViR.

# FAViR is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.

# FAViR is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.

# You should have received a copy of the GNU General Public License
# along with FAViR.  If not, see <http://www.gnu.org/licenses/>.

################################################################
# util.R - utility functions for the R actuarial example book

library(ChainLadder)

################### Generic utility functions

NumTrue <- function(vec) {
  # Count the number of TRUEs in vec
  return(length(vec[vec == TRUE]))
}

Assert <- function(boolean, err.msg=NULL) {
  # Abort and display traceback and err.msg if boolean is false
  #
  # Args:
  # boolean - TRUE or something other than TRUE
  # err.msg - additional message to display if boolean is not TRUE
  #
  # Output: If boolean is TRUE, do nothing.  Otherwise display err.msg
  # and print a traceback.
  if(isTRUE(boolean))
    return(invisible(NULL))
  call.list <- as.list(match.call())
  cat("----------------------------------------------------\n", file=stderr())
  cat("Error: Expression \"", as.character(as.expression(call.list[[2]])),
      "\"\n    evaluates to ", boolean, " instead of TRUE.\nTraceback:\n",
      sep="", file=stderr())
  for(i in 1:sys.nframe()) {
    trace.call <- sys.call(which=i)
    cat("    ", as.character(as.expression(trace.call)), "\n", sep="",
        file=stderr())
  }
  if(!is.null(err.msg))
    cat("Additional Comment:\n    ", err.msg, "\n",  sep="", file=stderr())
  cat("----------------------------------------------------\n", file=stderr())
  stop()
}

AssertSingle <- function(..., err.msg=NULL) {
  # Raise an error message if value does not have length 1
  # This is just a short-cut for Assert(length(value) == 1) for each
  # value in values.
  lengths <- sapply(list(...), length) # vector of lengths
  if(all(lengths == 1))
    return(NULL)
  bad.index <- which(lengths != 1)[1]
  bad.value <- list(...)[[bad.index]]
  bad.call <- as.list(match.call())[[1 + bad.index]]
  cat("----------------------------------------------------\n", file=stderr())
  cat("Error: Expression \"", as.character(as.expression(bad.call)),
      "\"\n    evaluates to ", as.character(bad.value), " and has length ",
      length(bad.value), " instead of 1.\nTraceback:\n", sep="", file=stderr())
  for(i in 1:sys.nframe()) {
    trace.call <- sys.call(which=i)
    cat("    ", as.character(as.expression(trace.call)), "\n", sep="",
        file=stderr())
  }
  if(!is.null(err.msg))
    cat("Additional Comment:\n    ", err.msg, "\n",  sep="", file=stderr())
  cat("----------------------------------------------------\n", file=stderr())
  stop()
}

StrSplit <- function(string, char) {
  # Split the string based on specified char, like Python's split
  #
  # Args:
  # string - the character string to split into multiple pieces
  # char - a single character which divides the piecs
  #
  # Output: a character vector containing the components separated by char
  AssertSingle(string, char)
  split <- strsplit(string, char, fixed=TRUE)[[1]]

  # Now need to make sure no trailing char
  l <- nchar(string)
  if (identical(substr(string, l, l), char))
    return(append(split, ""))
  else return(split)
}

StrSplitOnce <- function(string, char) {
  # Split string into two pieces: before and after char
  #
  # Args: string and char as with StrSplit
  #
  # Output: a character vector of length 2, where the components are
  # separated by char in the original string.  If there are multiple
  # chars in the original, split after the first
  split <- StrSplit(string, char)
  len <- length(split)
  if (len == 2)
    return(split)

  Assert(len >= 2)
  return(c(split[1], paste(split[2:len], collapse=char)))
}

Indicies <- function(vec) {
  # Return the indicies of vector vec
  len <- length(vec)
  if (len == 0)
    return(c())
  return(1:len)
}

RowIndicies <- function(x) {
  # Return the indicies of the rows of matrix or data frame x
  rows <- nrow(x)
  if (rows == 0)
    return(c())
  return(1:rows)
}

Last <- function(x, n=1, allow.fewer=FALSE) {
  # Return the last n elements of vector or data frame x
  Assert(n >= 0, "n should be a single non-negative integer")

  if (is.data.frame(x)) {
    if (n == 0)
      return(data.frame())
    len <- nrow(x)
    if (len < n) {
      Assert(allow.fewer, "Data frame is too short in call to Last()")
      return(x)
    }
    return(x[(len - n + 1):len, ])
  }
  
  if (n == 0)
    return(c())
  len <- length(x)
  if (len < n) {
    Assert(allow.fewer, "Vector is too short in call to Last()")
    return(x)
  }
  return(x[(len - n + 1):len])
}

LookupDF <- function(df, ..., allow.NULL = FALSE) {
  # Lookup and return the relevant rows from data frame df
  #
  # This is similar to vlookup in excel---only one row will be
  # returned.  if allow.NULL is true, if there are no matches NULL
  # will be returned.  Otherwise, signal error.
  lookup.args <- list(...)
  arg.names <- names(lookup.args)
  df.names <- names(df)
  Assert(all(arg.names %in% df.names), "Specified field missing from lookup df")
  Assert(length(arg.names) == length(unique(arg.names)))
  bool.vec <- rep(TRUE, nrow(df))

  for (name in names(lookup.args)) {
    arg.val <- lookup.args[[name]]
    AssertSingle(arg.val, "Vector not allowed inside field condition")
    bool.vec <- bool.vec & df[[name]] == arg.val
  }

  num.matches <- NumTrue(bool.vec)
  Assert(num.matches <= 1, "Too many rows match LookupDF conditions")
  if (num.matches == 0) {
    Assert(allow.NULL, "No LookupDF matches found")
    return(NULL)
  }
  return(df[bool.vec, ])
}

RowDF <- function(df, row.nums) {
  # Return a df representing the given row(s) of the data frame
  # This is because the basic indexing function [ ] seems buggy (!)
  # for instance, try
  # class(data.frame(a=1:4)[1, ])
  # and the result is not a data frame.
  #
  # Args: a data frame, and a vector of row indicies
  # Output: a data frame subset
  Assert("data.frame" %in% class(df), "Data frame required")
  len <- nrow(df)
  Assert(all(1 <= row.nums && row.nums <= len), "Row number out of range")
  Assert(length(df) > 0, "Data frame has 0 length")
  
  if (length(df) >= 2)
    # The normal subscripting seems to work OK here
    return(df[row.nums, ])
  result <- data.frame(df[[1]][row.nums])
  names(result) <- names(df)
  return(result)
}

LookupDFMult <- function(df, ..., allow.NULL=FALSE) {
  # Like LookupDF, but allow each argument to have multiple values
  # The output will be a DF with one row for each element in the args
  lookup.args <- data.frame(...)
  result <- data.frame()
  for (i in RowIndicies(lookup.args)) {
    single.lookup.args <- c(list(df=df),
                            as.list(RowDF(lookup.args, i)),
                            list(allow.NULL=allow.NULL))
    output <- do.call(LookupDF, single.lookup.args)
    result <- rbind(result, output)
  }
  return(result)
}

TransposeDF <- function(list.of.lists) {
  # Create data frame from a list of lists
  #
  # Each list should have the same names.  If the large list has
  # names, they will be used as the data frame row names.
  num.rows <- length(list.of.lists)
  if (identical(num.rows, 0))
    return(data.frame())

  first.list <- list.of.lists[[1]]
  Assert(length(first.list) >= 1, "First element of list.of.lists is empty")

  # Initialize df with right columns and row names
  return.df <- list()
  Assert(!any(is.null(names(first.list))), "Each list must have full names")
  for (name in names(first.list))
    return.df[[name]] <- rep(NA, num.rows)
  return.df <- as.data.frame(return.df)
  
  # Now populate row by row
  for (i in 1:num.rows) {
    cur.row <- list.of.lists[[i]]
    for (name in names(cur.row))
      return.df[i, name] <- cur.row[[name]]
  }

  # Add row names and return
  rownames(return.df) <- names(list.of.lists)
  return(return.df)
}

RemoveNARows <- function(df) {
  # Remove any row with an NA from a data frame and return result
  for (i in Indicies(df))
    df <- df[!is.na(df[[i]]), ]
  return(df)
}

FavirLoad <- function(filename) {
  # Source the given file and return a new favir.environment
  e <- new.env()
  evalq(source(filename, local=TRUE), e)
  return(structure(list(e), class="favir.environment"))
}

`$.favir.environment` <- function(fe, varname, ...) {
  # Return an element from a favir.environment
  return(get(varname, fe[[1]], inherits=FALSE))
}

with.favir.environment <- function(fe, expr, ...) {
  # Run with inside a favir environment (e.g. "with(fe, ls())")
  eval(expr, fe[[1]])
}

################### Example Input Data -- Replace this with your own

incurred.devel.df <- data.frame(origin = c(1995, 1996, 1997, 1998,
                                  1999, 2000, 2001, 2002, 2003, 2004,
                                  2005, 2006, 1995, 1996, 1997, 1998,
                                  1999, 2000, 2001, 2002, 2003, 2004,
                                  2005, 1995, 1996, 1997, 1998, 1999,
                                  2000, 2001, 2002, 2003, 2004, 1995,
                                  1996, 1997, 1998, 1999, 2000, 2001,
                                  2002, 2003, 1995, 1996, 1997, 1998,
                                  1999, 2000, 2001, 2002, 1995, 1996,
                                  1997, 1998, 1999, 2000, 2001, 1995,
                                  1996, 1997, 1998, 1999, 2000, 1995,
                                  1996, 1997, 1998, 1999, 1995, 1996,
                                  1997, 1998, 1995, 1996, 1997, 1995,
                                  1996, 1995),
                                dev = c(3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
                                  3, 3, 15, 15, 15, 15, 15, 15, 15,
                                  15, 15, 15, 15, 27, 27, 27, 27, 27,
                                  27, 27, 27, 27, 27, 39, 39, 39, 39,
                                  39, 39, 39, 39, 39, 51, 51, 51, 51,
                                  51, 51, 51, 51, 63, 63, 63, 63, 63,
                                  63, 63, 75, 75, 75, 75, 75, 75, 87,
                                  87, 87, 87, 87, 99, 99, 99, 99, 111,
                                  111, 111, 123, 123, 135),
                                value = c(44, 42, 17, 10, 13, 2, 4, 2,
                                  3, 4, 21, 13, 1331, 1244, 1088, 781,
                                  937, 751, 1286, 911, 1398, 1130,
                                  915, 3319, 3508, 3438, 3135, 3506,
                                  2639, 3570, 5023, 4021, 3981, 4020,
                                  4603, 4169, 4085, 4828, 3622, 4915,
                                  6617, 4825, 4232, 4842, 4371, 4442,
                                  5447, 3931, 5377, 7194, 4252, 4970,
                                  4482, 4777, 5790, 4077, 5546, 4334,
                                  5059, 4626, 4914, 6112, 4244, 4369,
                                  5083, 4734, 5110, 6295, 4386, 5155,
                                  4794, 5176, 4395, 5205, 4804, 4401,
                                  5205, 4399))

## Another possible incurred triangle
## incurred.devel.df <- data.frame(origin = c(1981, 1982, 1983, 1984,
##                                   1985, 1986, 1987, 1988, 1989, 1990,
##                                   1981, 1982, 1983, 1984, 1985, 1986,
##                                   1987, 1988, 1989, 1981, 1982, 1983,
##                                   1984, 1985, 1986, 1987, 1988, 1981,
##                                   1982, 1983, 1984, 1985, 1986, 1987,
##                                   1981, 1982, 1983, 1984, 1985, 1986,
##                                   1981, 1982, 1983, 1984, 1985, 1981,
##                                   1982, 1983, 1984, 1981, 1982, 1983,
##                                   1981, 1982, 1981),
##                                 dev = c(9, 9, 9, 9, 9, 9, 9, 9, 9,
##                                   9, 21, 21, 21, 21, 21, 21, 21, 21,
##                                   21, 33, 33, 33, 33, 33, 33, 33, 33,
##                                   45, 45, 45, 45, 45, 45, 45, 57, 57,
##                                   57, 57, 57, 57, 69, 69, 69, 69, 69,
##                                   81, 81, 81, 81, 93, 93, 93, 105,
##                                   105, 117 ),
##                                 value = c(5012L, 106L, 3410L, 5655L,
##                                   1092L, 1513L, 557L, 1351L, 3133L,
##                                   2063L, 8269L, 4285L, 8992L, 11555L,
##                                   9565L, 6445L, 4020L, 6947L, 5395L,
##                                   10907L, 5396L, 13873L, 15766L,
##                                   15836L, 11702L, 10946L, 13112L,
##                                   11805L, 10666L, 16141L, 21266L,
##                                   22169L, 12935L, 12314L, 13539L,
##                                   13782L, 18735L, 23425L, 25955L,
##                                   15852L, 16181L, 15599L, 22214L,
##                                   26083L, 26180L, 18009L, 15496L,
##                                   22863L, 27067L, 18608L, 16169L,
##                                   23466L, 18662L, 16704L, 18834L))
incurred.devel.tri <- as.triangle(incurred.devel.df)
incurred.devel.tri.df <- cbind(data.frame(origin=1995:2006),
                               as.data.frame(unclass(incurred.devel.tri)))

paid.devel.df <- data.frame(origin = c(1995, 1996, 1997, 1998, 1999,
                              2000, 2001, 2002, 2003, 2004, 2005,
                              2006, 1995, 1996, 1997, 1998, 1999,
                              2000, 2001, 2002, 2003, 2004, 2005,
                              1995, 1996, 1997, 1998, 1999, 2000,
                              2001, 2002, 2003, 2004, 1995, 1996,
                              1997, 1998, 1999, 2000, 2001, 2002,
                              2003, 1995, 1996, 1997, 1998, 1999,
                              2000, 2001, 2002, 1995, 1996, 1997,
                              1998, 1999, 2000, 2001, 1995, 1996,
                              1997, 1998, 1999, 2000, 1995, 1996,
                              1997, 1998, 1999, 1995, 1996, 1997,
                              1998, 1995, 1996, 1997, 1995, 1996,
                              1995),
                            dev = c(3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
                              3, 15, 15, 15, 15, 15, 15, 15, 15, 15,
                              15, 15, 27, 27, 27, 27, 27, 27, 27, 27,
                              27, 27, 39, 39, 39, 39, 39, 39, 39, 39,
                              39, 51, 51, 51, 51, 51, 51, 51, 51, 63,
                              63, 63, 63, 63, 63, 63, 75, 75, 75, 75,
                              75, 75, 87, 87, 87, 87, 87, 99, 99, 99,
                              99, 111, 111, 111, 123, 123, 135),
                            value = c(3, 1, 1, 1, 1, 1, 1, 1, 1, 4, 1,
                              1, 503, 465, 534, 329, 399, 328, 443,
                              401, 326, 524, 323, 2474, 2621, 2541,
                              2204, 2496, 1849, 2566, 3078, 2372,
                              2784, 3719, 4122, 3807, 3673, 4304,
                              3124, 4208, 5459, 4132, 4094, 4618,
                              4192, 4242, 5197, 3693, 5074, 6748,
                              4194, 4882, 4374, 4616, 5674, 3966,
                              5474, 4303, 4997, 4544, 4827, 6031,
                              4164, 4350, 5041, 4679, 5051, 6244,
                              4382, 5111, 4761, 5145, 4394, 5172,
                              4787, 4394, 5191, 4398))
paid.devel.tri <- as.triangle(paid.devel.df)
paid.devel.tri.df <- cbind(data.frame(origin=1995:2006),
                           as.data.frame(unclass(paid.devel.tri)))

premium.df <- data.frame(origin=1995:2006, premium=rep(6000, 12))
premium.df$apriori.loss <- 0.8*premium.df$premium
premium.df$apriori.ratio <- premium.df$apriori.loss / premium.df$premium

################### LDF Selection functions

LDF.GetAgeToAge <- function(devel.df) {
  # Return dataframe of all age-to-age LDFs
  #
  # Arg: development dataframe with origin, dev, and value fields
  #
  # Output: Dataframe of LDFs with fields origin, start.dev, end.dev,
  # and LDF.
  sorted.df <- devel.df[order(devel.df$origin, devel.df$dev), ]
  len <- nrow(sorted.df)
  Assert(len >= 2, "devel.df has too few rows")
  
  result <- data.frame(origin=c(), start.dev=c(), end.dev=c(), LDF=c())
  for (i in 1:(len - 1)) {
    cur.row <- sorted.df[i, ]
    next.row <- sorted.df[i+1, ]

    if (cur.row$origin != next.row$origin)
      next
    if (cur.row$value == 0 || is.na(cur.row$value) || is.na(next.row$value))
      next
    LDF <- next.row$value / cur.row$value
    result <- rbind(result, list(origin=cur.row$origin, start.dev=cur.row$dev,
                                 end.dev=next.row$dev, LDF=LDF))
  }
  return(result)
}
  
LDF.GetAgeToAgeMatrix <- function(age.df) {
  # Return matrix of AgeToAge factors
  #
  # Arg: age-to-age data frame as produced by LDF.GetAgeToAge
  # Output: "formatted" matrix of age-to-age factors
  origins <- sort(unique(age.df$origin))
  dev.df <- unique(data.frame(from=age.df$start.dev, to=age.df$end.dev))
  def.df <- dev.df[order(dev.df$from, dev.df$to), ]

  result <- matrix(nrow=length(origins), ncol=nrow(dev.df))
  rownames(result) <- origins
  colnames(result) <- paste(dev.df$from, "to", dev.df$to)
  for (i in Indicies(origins)) {
    origin <- origins[i]
    for (j in Indicies(colnames(result))) {
      LDF <- LookupDF(age.df, origin=origin, start.dev=dev.df$from[j],
                      end.dev=dev.df$to[j], allow.NULL=TRUE)$LDF
      if (!is.null(LDF))
        result[i, j] <- LDF
    }
  }
  return(result)
}

LDF.getAverages <- function(triangle.df, LDF.df) {
  # Return DF showing different ways of averaging the LDFs
  #
  # Args: Two data frames, one with raw losses, and the other of LDFs,
  # like the kind produced by LDF.GetAgeToAge
  #
  # Output: A data frame with the fields method (char), start.dev
  # (number), end.dev (number), and LDF (number)
  GetLDFs <- function(start.dev, end.dev) {
    # Get data frame of matching LDFs from LDF.df, sorted by origin
    sub.df <- LDF.df[LDF.df$start.dev == start.dev &
                     LDF.df$end.dev == end.dev, ]
    return(sub.df[order(sub.df$origin), ])
  }

  GetOrigins <- function(start.dev, end.dev) {
    # Return DF of start.value, end.value matching given start and end devs
    start.df <- triangle.df[triangle.df$dev == start.dev, ]
    end.df <- triangle.df[triangle.df$dev == end.dev, ]
    joined.df <- merge(start.df, end.df, by="origin")
    pruned.df <- data.frame(origin=joined.df$origin,
                            start.value=joined.df$value.x,
                            end.value=joined.df$value.y)
    return(pruned.df[order(pruned.df$origin), ])
  }
  
  AvgLast <- function(start.dev, end.dev, n) {
    # Return straight average of last n LDFs
    sub.df <- GetLDFs(start.dev, end.dev)
    num.rows <- nrow(sub.df)
    if (num.rows <= n)
      return(mean(sub.df$LDF))
    else return(mean(sub.df$LDF[(num.rows - n + 1):num.rows]))
  }

  AvgXHiLo <- function(start.dev, end.dev) {
    # Return average, skipping the high and low LDFs (if at least 3)
    LDFs <- sort(GetLDFs(start.dev, end.dev)$LDF)
    if (length(LDFs) >= 3)
      return(mean(LDFs[2:(length(LDFs) - 1)]))
    return(mean(LDFs))
  }
  
  AvgWeighted <- function(start.dev, end.dev, n) {
    # Return weighted average of last n LDFs
    sub.df <- GetOrigins(start.dev, end.dev)
    num.rows <- nrow(sub.df)
    if (nrow(sub.df) > n)
      sub.df <- sub.df[(num.rows - n + 1):num.rows, ]
    if (sum(sub.df$start.value) == 0)
      return(NA)
    return(sum(sub.df$end.value) / sum(sub.df$start.value))
  }

  interval.df <- unique(data.frame(start.dev=LDF.df$start.dev,
                                   end.dev=LDF.df$end.dev))
  result.df <- data.frame(method=c(), start.dev=c(), end.dev=c(), LDF=c())
  for (i in RowIndicies(interval.df)) {
    start.dev <- interval.df$start.dev[i]
    end.dev <- interval.df$end.dev[i]
    avg <- data.frame(method="Average", start.dev=start.dev,
                      end.dev=end.dev, LDF=AvgLast(start.dev, end.dev, Inf))
    avg.xhilo <- data.frame(method="Avg xHi,Lo", start.dev=start.dev,
                            end.dev=end.dev, LDF=AvgXHiLo(start.dev, end.dev))
    avg.last5 <- data.frame(method="Avg Last 5", start.dev=start.dev,
                            end.dev=end.dev,
                            LDF=AvgLast(start.dev, end.dev, 5))
    avg.weighted <- data.frame(method="Weighted Avg",
                               start.dev=start.dev, end.dev=end.dev,
                               LDF=AvgWeighted(start.dev, end.dev, Inf))
    avg.wlast5 <- data.frame(method="Weighted Last 5",
                             start.dev=start.dev, end.dev=end.dev,
                             LDF=AvgWeighted(start.dev, end.dev, 5))
    result.df <- rbind(result.df, avg, avg.xhilo, avg.last5, avg.weighted,
                       avg.wlast5)
  }
  return(result.df)
}

LDF.getAvgMatrix <- function(avg.df) {
  # Return matrix of LDFs averaged in various ways
  #
  # Arg: a data frame of averages, as given by LDF.getAverages
  #
  # Output: a labeled matrix of averages
  methods <- unique(avg.df$method)
  dev.df <- unique(data.frame(from=avg.df$start.dev, to=avg.df$end.dev))
  dev.df <- dev.df[order(dev.df$from, dev.df$to), ]

  result <- matrix(nrow=length(methods), ncol=nrow(dev.df))
  rownames(result) <- methods
  colnames(result) <- paste(dev.df$from, "to", dev.df$to)

  for (i in Indicies(methods)) {
    method <- methods[i]
    for (j in Indicies(colnames(result))) {
      avg <- LookupDF(avg.df, method=method, start.dev=dev.df$from[j],
                      end.dev=dev.df$to[j], allow.NULL=TRUE)$LDF
      if (!is.null(avg))
        result[i, j] <- avg
    }
  }
  return(result)
}

LDF.AgeToUlt <- function(age.to.age.df, tail.factor) {
  # Return data frame of LDFs to ultimate
  #
  # Args: a data frame of age to age factors, as produced by
  # LDF.getAgeToAge, and a tail factor to ultimate
  #
  # Output: a data frame with fields dev (age) and LDF (to ultimate)
  Assert(length(unique(age.to.age.df$start.dev)) == nrow(age.to.age.df))
  sorted.df <- age.to.age.df[order(-age.to.age.df$start.dev), ]
  df <- data.frame(dev=c(sorted.df$end.dev[1], sorted.df$start.dev),
                   LDF=cumprod(c(tail.factor, sorted.df$LDF)))
  return(df[order(df$dev), ])
}

LDF.AgeToUltFDF <- function(age.to.ult.df, label="LDFs to Ultimate",
                            textsize="normalsize",
                            caption="Selected LDFs to Ultimate") {
  # Return printable display of final age to ultimate LDFs
  l <- c(list("LDFs to Ultimate"), as.list(age.to.ult.df$LDF))
  fdf <- Favir.data.frame(as.data.frame(l), label=label, caption=caption)
  names(fdf) <- c("X", as.character(age.to.ult.df$dev))
  headings <- as.character(age.to.ult.df$dev)
  names(headings) <- as.character(age.to.ult.df$dev)
  FieldHeadings(fdf) <- c(list(X=""), headings)
  FieldGroup(fdf, "LDF") <- 2:length(fdf)
  GroupHeading(fdf, "LDF") <- "Development Age"
  TextSize(fdf) <- textsize
  return(fdf)
}

###################### Tail Methods
  
Tail.McClenahan <- function(LDFs) {
  # Compute tail method according to McClenahan's method
  #
  # Args: LDFs, a vector of age to age devel factors
  #
  # Output: single age to ultimate tail factor computed using
  # McClenahan's method
  LDFs <- pmax(LDFs, 1.0)
  if (all(LDFs <= 1.000001))
    return(1.0)
  
  # This method doesn't work well with LDFs near zero, so bump them up
  LDFs <- ifelse(LDFs <= 1.000001, 1.000001, LDFs)

  cumulative.loss <- cumprod(c(1.0, LDFs))
  inc.loss <- (cumulative.loss[2:length(cumulative.loss)] -
               cumulative.loss[1:(length(cumulative.loss) - 1)])
  tail.model <- lm(log(inc.loss) ~ Indicies(inc.loss))
  r <- exp(coef(tail.model)[2])
  remaining.loss <- sum(inc.loss[length(inc.loss)]*r^(1:100))
  return(1 + remaining.loss / Last(cumulative.loss))
}

Tail.ModMcClenahan <- function(devs, cumulative) {
  # Get tail function according to modified McClenahan's method
  #
  # This is a modified version of McClenahan's function.  Unlike some
  # tail functions, it allows the date of loss developments to be
  # given explicitly, and returns a function extrapolating future
  # payments.
  #
  # Args: devs, a vector of ages (e.g. accident year age in months)
  # cumulative, a vector of cumulative payments at each age
  #
  # Output: a function from development age to cumulative loss.
  Assert(length(devs) >= 2 && length(devs) == length(cumulative))
  len <- length(devs)
  df <- data.frame(dev=devs, cumulative=cumulative)
  start <- c(Last(df$cumulative), 1/Last(df$dev))
  names(start) <- c("ult", "r")
  cum.model <- nls(cumulative ~ ult * (1 - exp(-r*dev)), df, start=start)

  # Fit to last payment
  r <- coef(cum.model)["r"]
  ult <- coef(cum.model)["ult"]
  pred.last <- ult * (1 - exp(-r*Last(devs)))
  actual.last <- Last(cumulative)
  
  # Now construct tail function
  tail.func <- function(dev) {
    # Return cumulative paid to date given dev
    return(ult * actual.last / pred.last * (1 - exp(-r*dev)))
  }
  return(tail.func)
}

Tail.ExponentialDecay <- function(LDFs) {
  # Compute tail method by the exponential decay method, with Boor's
  # "improvement 2"
  #
  # Args: LDFs, a vector of age to age devel factors
  #
  # Output: single age to ultimate tail factor computed by assuming
  # exponential decay of LDFs.
  LDFs <- pmax(LDFs, 1.0)
  if (all(LDFs <= 1.000001))
    return(1.0)
  
  # This method doesn't work well with LDFs near zero, so bump them up
  LDFs <- ifelse(LDFs <= 1.000001, 1.000001, LDFs)
  tail.model <- lm(log(LDFs - 1) ~ Indicies(LDFs))
  r <- exp(coef(tail.model)[2])
  future.LDFs <- 1 + (Last(LDFs)-1)*r^(1:100)
  return(prod(future.LDFs))
}

Tail.Sherman <- function(LDFs, future.links=15) {
  # Compute tail method by Sherman's method
  #
  # Args: LDFs, a vector of age to age devel factors
  # future.links, the number of years to extrapolate to
  #
  # Output: single age to ultimate tail factor computed by assuming
  # power law governs LDF decay
  LDFs <- pmax(LDFs, 1.0)
  if (all(LDFs <= 1.000001))
    return(1.0)
  # This method doesn't work well with LDFs near zero, so bump them up
  LDFs <- ifelse(LDFs <= 1.000001, 1.000001, LDFs)

  log.year <- log(Indicies(LDFs))
  tail.model <- lm(log(LDFs - 1) ~ log.year)
  future.indep <- data.frame(log.year=log(Last(Indicies(LDFs)) +
                               0:future.links))
  future.LDFs <- 1 + exp(predict(tail.model, newdata=future.indep))
  # Now fit future.LDFs to last link and cap at 1.0
  tail <- (prod(future.LDFs[2:length(future.LDFs)]) *
           (Last(LDFs) / future.LDFs[1]))
  return(max(tail, 1.0))
}

Tail.Summary <- function(LDF.age.to.age.df) {
  # Return matrix summarizing tail results
  #
  # Args: data frame with fields start.dev, end.dev, and LDF
  # Output: matrix with one column and named rows
  LDFs <- LDF.age.to.age.df$LDF
  mcclenahan <- Tail.McClenahan(LDFs)
  ModMcc <- Tail.ModMcClenahan(LDF.age.to.age.df$end.dev, cumprod(LDFs))
  mod.mcc.tail <- ModMcc(1e6) / ModMcc(Last(LDF.age.to.age.df$end.dev))
  expon <- Tail.ExponentialDecay(LDFs)
  sherman <- Tail.Sherman(LDFs)
  result <- matrix(c(mcclenahan, mod.mcc.tail, expon, sherman), nrow=4)
  colnames(result) <- "Tail Factor to Ultimate"
  rownames(result) <- c("McClenahan Method (exponential)",
                        "Modified McClenahan Method",
                        "Exponential Decay of LDFs to 1.0",
                        "Sherman Method (inverse power law)")
  return(result)
}
  

############################# LDF Interpolation

LDF.MakeInterpolate <- function(LDFs, devs, TailCumulative) {
  # Use simple algorithm to interpolate LDFs
  #
  # Input: vectors of LDFs to ultimate and development ages (same length)
  # Also takes a tail function that computes cumulative at dev.
  #
  # Output: Function from development age(s) to LDF to ultimate(s).
  Assert(length(LDFs) >= 2 && length(LDFs) == length(devs))

  # Make tail function
  tail.scale <- Last(LDFs) * TailCumulative(Last(devs))
  TailLDFs <- function(dev) tail.scale / TailCumulative(dev)
  
  # Make head function
  power.b <- log((LDFs[1] - 1) / (LDFs[2] - 1)) / log(devs[1] / devs[2])
  power.a <- (LDFs[1] - 1) / devs[1] ^ power.b
  head.func <- function(dev) (power.a * dev ^ power.b) + 1

  # For body, add extra point at head and tail for smoothness
  prefix <- min(devs) * .99
  suffix <- max(devs) * 1.01
  body.func <- splinefun(c(prefix, devs, suffix),
                         c(head.func(prefix), LDFs, TailLDFs(suffix)))
  
  InterpolateLDF <- function(dev.vec) {
    # Return interpolated LDF given development age
    return(ifelse(dev.vec < min(devs), head.func(dev.vec),
                  ifelse(dev.vec > max(devs), TailLDFs(dev.vec),
                         body.func(dev.vec))))
  }
  return(InterpolateLDF)
}

                            
LDF.getUlt <- function(LDF.selected) {
  # Return matrix summarizing selected LDFs
  LDF.ultimate <- LDF.selected
  for (i in (length(LDF.ultimate)-1):1)
    LDF.ultimate[i] <- LDF.ultimate[i+1] * LDF.selected[i]
  pct.reported <- 100/LDF.ultimate

  result <- rbind(LDF.selected, LDF.ultimate, pct.reported)
  rownames(result) <- c('Selected LDFs', 'LDFs to Ult', '% Reported')
  return (result)
}

Latest <- function(df.tri) {
  # Return the latest diagonal in data frame format from a data frame triangle
  helper <- function(origin) {
    # Return dev and value as list for last eval given origin
    sub.df <- df.tri[df.tri$origin==origin, ]
    if (nrow(sub.df) < 1)
      return(list(origin=NA, dev=NA, value=NA))
    return(LookupDF(df.tri, origin=origin, dev=max(sub.df$dev)))
  }
  origins <- unique(df.tri$origin)
  return(TransposeDF(Map(helper, origins)))
}



######################### Basic method functions

ChainladderFDF <- function(latest.df, LDF.selected.ult) {
  # Return a favir data frame indicating ultimate for each origin
  # using Chain Ladder on
  result <- latest.df
  result$ult.ldf <- LookupDFMult(LDF.selected.ult,
                                         dev=latest.df$dev)$LDF
  result$pct.devel <- 1 / result$ult.ldf
  result$ult <- result$value * result$ult.ldf
  result <- Favir.data.frame(result)
  FieldHeadings(result) <- list(dev="Development Age", value="Latest Diagonal",
                                ult.ldf="LDF to Ultimate",
                                pct.devel="Percent Developed",
                                ult="Ultimate Loss")
  FieldGroup(result, "money") <- c("value", "ult")
  FieldGroup(result, "origin") <- "origin"
  FieldGroup(result, "dev") <- "dev"
  FieldGroup(result, "LDF") <- "ult.ldf"
  FieldGroup(result, "percent") <- "pct.devel"
  return(result)
}

Bornhuetter <- function(LDFs, latest, apriori) {
  # Return a vector of Bornhuetter-Ferguson developed ultimate losses
  #
  # Args: LDFs: a vector of LDFs to ultimate
  # latest: a vector of latest loss valuations matching LDFs
  # apriori: a vector of pre-data loss estimates
  Assert(length(LDFs) == length(latest) && length(latest) == length(apriori))
  return((1 - 1/LDFs) * apriori + latest)
}  

BornhuetterFDF <- function(latest.df, LDF.selected.ult, apriori.byyear) {
  # Return a favir data frame indicating ultimate for each origin
  # using Bornhuetter-Ferguson
  df <- latest.df
  df$ult.ldf <- LookupDFMult(LDF.selected.ult, dev=latest.df$dev)$LDF
  df$pct.devel <- 1 / df$ult.ldf
  df$a.priori <- LookupDFMult(apriori.byyear,
                              origin=latest.df$origin)$apriori.loss
  df$ult <- Bornhuetter(df$ult.ldf, df$value, df$a.priori)

  fdf <- Favir.data.frame(df)
  FieldHeadings(fdf) <- list(dev="Development Age", value="Latest Diagonal",
                             ult.ldf="LDF to Ultimate",
                             pct.devel="Percent Developed",
                             a.priori="A Priori Loss",
                             ult="BF Ultimate Loss")
  FieldGroup(fdf, "origin") <- "origin"
  FieldGroup(fdf, "money") <- c("value", "a.priori", "ult")
  FieldGroup(fdf, "dev") <- "dev"
  FieldGroup(fdf, "percent") <- "pct.devel"
  return(fdf)
}

CapeCod <- function(latest.df, LDF.selected.ult, premium.byyear) {
  # Run CapeCod method to compute ultimate losses
  #
  # Args: Two data frames, one that has the latest evaluations by year,
  # another with matching LDFs
  #
  # Output: a favir data frame with the fields
  # premium - the premium used for the computation
  # used.premium - the Cape Cod "used up premium" for each year
  # exp.ratio - the ratio of loss to used premium
  df <- data.frame(origin=latest.df$origin, value=latest.df$value)
  df$LDF <- LookupDFMult(LDF.selected.ult, dev=latest.df$dev)$LDF
  df$premium <- LookupDFMult(premium.byyear, origin=latest.df$origin)$premium
  df$used.premium <- df$premium / df$LDF
  df$exp.ratio <- df$value / df$used.premium

  fdf <- Favir.data.frame(df)
  FieldHeadings(fdf) <- list(value="Latest Diagonal", LDF="LDF to Ultimate",
                             premium="Total Premium",
                             used.premium="Used-Up Premium",
                             exp.ratio="Expected Loss Ratio")
  FieldGroup(fdf, "origin") <- "origin"
  FieldGroup(fdf, "money") <- c("value", "premium", "used.premium")
  FieldGroup(fdf, "percent") <- "exp.ratio"
  FieldGroup(fdf, "LDF") <- "LDF"
  return(fdf)
}

PrintCombinedTable <- function(triangle.fdf, long.df, label="", caption="",
                               orientation="sideways", textsize=NULL) {
  # Print the combined three-layer LDF table
  #
  # Args:
  # triangle.fdf - a favir data frame
  # long.df - the same triangle in long form
  # label, caption, textsize, orientation - see arguments to Favir.data.frame
  #
  # Output: A data frame of LDFs averaged in various ways
  # It also prints the latex of the combined table.
  combined.table.head <- triangle.fdf
  Label(combined.table.head) <- label
  Caption(combined.table.head) <- caption
  if (!is.null(textsize))
    TextSize(combined.table.head) <- textsize

  LDF.age.to.age <- LDF.GetAgeToAge(long.df)
  LDF.age.matrix <- LDF.GetAgeToAgeMatrix(LDF.age.to.age)
  LDF.age.fdf <- MatrixToFDF(LDF.age.matrix, label="Simple LDFs",
                             rowname.col="origin")
  FieldHeadings(LDF.age.fdf) <- list(origin=origin.field.heading)
  FieldGroup(LDF.age.fdf, "LDF") <- 2:length(LDF.age.fdf)
  GroupHeading(LDF.age.fdf, "LDF") <- "Age to Age Loss Development Factors"

  LDF.avg.df <- LDF.getAverages(long.df, LDF.age.to.age)
  LDF.avg.fdf <- MatrixToFDF(LDF.getAvgMatrix(LDF.avg.df), rowname.col="method",
                             orientation=orientation, label="Average LDFs")
  FieldHeadings(LDF.avg.fdf) <- list(method="")
  FieldGroup(LDF.avg.fdf, "LDF") <- 2:length(LDF.age.fdf)
  GroupHeading(LDF.avg.fdf, "LDF") <- "Averaged Age-to-Age LDFs"

  print(RBindPreLatex(PreLatex(combined.table.head), PreLatex(LDF.age.fdf),
                      PreLatex(LDF.avg.fdf)))
  return(LDF.avg.df)
}

MakeSummaryFDFs <- function(summary.df, origin.field.heading) {
  # Return summary FDFs, possibly splitting them between multiple tables
  maxcols <- 5
  summary.origins <- unique(summary.df$origin)
  summary.methods <- unique(summary.df$method)
  origin.groups <- split(summary.origins,
                         (Indicies(summary.origins) - 1) %/% maxcols)

  MakeFDF <- function(origins) {
    # Return FDF with columns for the origins given
    subdf <- summary.df[summary.df$origin %in% origins, ]
    summary.wide <- reshape::recast(subdf, method ~ origin,
                                    id.var=c("method", "origin"))
    Assert(nrow(summary.wide) == length(summary.methods))
    Assert(ncol(summary.wide) >= 2 && ncol(summary.wide) <= maxcols + 1)
    fdf <- Favir.data.frame(summary.wide, label="summary",
                             caption="Multi-method Development Summary")
    FieldGroup(fdf, "money") <- 2:length(fdf)
    FieldHeadings(fdf) <- list(method="Method")
    if (Last(names(fdf)) == "Total" && length(fdf) > 2) {
      FieldGroup(fdf, "origins") <- 2:(length(fdf) - 1)
      GroupHeading(fdf, "origins") <- paste("Ultimate by", origin.field.heading)
    } else if (Last(names(fdf)) != "Total") {
      FieldGroup(fdf, "origins") <- 2:length(fdf)
      GroupHeading(fdf, "origins") <- paste("Ultimate by", origin.field.heading)
    }
    return(fdf)
  }

  return(lapply(origin.groups, MakeFDF))
}

################################## Functions for regression reserving

MakeDevModels <- function(dev, mack.lm) {
  # Given the mack lm model, produce models with intercepts
  #
  # Args:
  # dev - development period name
  # mack.lm - the lm object produced by the MackChainLadder function
  #
  # Output: A list with three elements
  # link.lm - the default mack model
  # intercept.lm - a lm object with just an intercept term
  # both.lm - model with both intercept and link
  df <- data.frame(increment=mack.lm$model$y - mack.lm$model$x,
                   start=mack.lm$model$x, weights=mack.lm$weights)
  return(list(dev=dev,
              link.lm=lm(increment ~ start + 0, weights = weights, data=df),
              intercept.lm=lm(increment ~ 1, weights = weights, data=df),
              both.lm=lm(increment ~ start, weights = weights, data=df)))
}

GetLinkDF <- function(dev, link.lm, intercept.lm, both.lm) {
  # Return data frame about fitting a single development period
  #
  # Args:
  # dev - development period as a constant
  # link.lm - the lm object produced by the MackChainLadder function
  # intercept.lm - lm object with just an intercept term
  # both.lm - lm object with both link and intercept
  #
  # Output: a data frame with these fields:
  # dev - the development period description
  # start - the initial loss amount
  # actual - the actual ending loss amount
  # link.only - the fitted final loss amounts using the link-only model
  # both - fitted losses from model with intercept and link
  # intercept.only - fitted losses from model with just intercept
  starting.loss <- link.lm$model$start
  predict.df <- data.frame(start=c(0, starting.loss))

  df <- data.frame(dev=dev, start=c(0, starting.loss),
                   actual=c(NA, link.lm$model$increment))
  df$link.only <- predict(link.lm, predict.df)
  df$intercept.only <- predict(intercept.lm, predict.df)
  df$both <- predict(both.lm, predict.df)
  return(df)
}

DevModelStats <- function(dev.model.list) {
  # Return a list of statistics from a model list as produced by MakeDevModels
  link.summary <- summary(dev.model.list$link.lm)
  intercept.summary <- summary(dev.model.list$intercept.lm)
  both.summary <- summary(dev.model.list$both.lm)
  t.values <- both.summary$coefficients[, "t value"]
  powers <- both.summary$coefficients[, "Pr(>|t|)"]
  link.RS <- 1 - (link.summary$sigma / intercept.summary$sigma)^2
  
  return(list(dev.period=dev.model.list$dev, link.RS=link.RS,
              both.RS=both.summary$r.squared,
              link.t=t.values[2], intercept.t=t.values[1], 
              link.power=powers[2], intercept.power=powers[1]))
}

FullTriangleDF <- function(half.tri, full.tri) {
  # Return a full triangle data frame with 0's
  #
  # Args:
  # half.tri - a half triangle in matrix form, with NAs where there are
  #            unobserved elements
  #
  # full.tri - a fully developed triangle.  The "dev" column does not
  #            have to match the half.tri.
  #
  # Output: A data frame with the columns origin, dev, value, and
  # actual.  the dev column will be taken from half.tri.  actual will
  # be TRUE or FALSE depending on whether the value was in the half
  # triangle.  Additionally, the observed part will be offset
  # backwards one period for plotting.
  half.df <- as.data.frame(half.tri)
  full.df <- as.data.frame(full.tri)
  Assert("triangle" %in% class(half.tri) && "triangle" %in% class(full.tri),
         "Must call FullTriangleDF with two triangle objects")
  Assert(nrow(half.df) == nrow(full.df), 
       "Error in triangle manipulation---number of rows don't match")
  Assert(all(names(full.df) == c("origin", "dev", "value")))
  
  # Sort them both identically
  half.df <- half.df[order(half.df$dev, half.df$origin), ]
  full.df <- full.df[order(full.df$dev, full.df$origin), ]
  full.df$dev <- half.df$dev # in case dev ages have been overwritten
  full.df <- rbind(data.frame(origin=unique(full.df$origin), dev=0, value=0),
                   full.df)

  dev.periods <- unique(full.df$dev)
  GetObserved <- function(origin, dev) {
    # Return TRUE or FALSE depending on whether the next development
    # period after dev has been observed
    if (dev == max(dev.periods))
      return(FALSE)
    next.dev <- dev.periods[min(which(dev.periods > dev))]
    Assert(length(next.dev) == 1 && next.dev > dev)
    
    next.value <- LookupDF(half.df, origin=origin, dev=next.dev)$value
    return(!is.na(next.value))
  }

  # Now update with "observed" column
  full.df$actual <- mapply(GetObserved, origin=full.df$origin, dev=full.df$dev)
  return(full.df)
}

MackSummaryFDF <- function(mack.model, triangle.df, origin.field.heading) {
  # Return an favir data frame summarizing the results of given mack model
  mack.byorigin <- summary(mack.model)$ByOrigin
  Assert(nrow(mack.byorigin) == nrow(triangle.df),
         "Number of rows in Mack Output doesn't match")
  new.col <- data.frame(origin=triangle.df$origin)
  mack.fdf <- Favir.data.frame(cbind(new.col, mack.byorigin))
  FieldHeadings(mack.fdf) <- list(origin=origin.field.heading,
                                  Latest="Latest Diagonal",
                                  Dev.To.Date="Percent Developed",
                                  Ultimate="Mack Ultimate",
                                  IBNR="Bulk Reserve",
                                  Mack.S.E="Mack Standard Error",
                                  `CV(IBNR)`="CV of Bulk Reserves")
  FieldGroup(mack.fdf, "money") <- c("Latest", "Ultimate", "IBNR",
                                     "Mack.S.E")
  FieldGroup(mack.fdf, "percent") <- "Dev.To.Date"
  FieldGroup(mack.fdf, "origin") <- "origin"
  mack.totals <- summary(mack.model)$Totals
  SummaryRow(mack.fdf) <- list(origin="Total",
                               Latest=mack.totals["Latest:", ],
                               Ultimate=mack.totals["Ultimate:", ],
                               IBNR=mack.totals["IBNR:", ],
                               Mack.S.E=mack.totals["Mack S.E.:", ],
                               `CV(IBNR)`=mack.totals["CV(IBNR):", ])
  return(mack.fdf)
}

MackResidualPlots <- function(mack.model, triangle.long) {
  # Make Mack model residual plots
  #
  # Args:
  # mack.model - the MackChainLadder object
  # triangle.long - a triangle in long data frame format.
  #
  # Output:
  # a list of models with names "loss", "origin", "cal", and "dev"
  resid <- residuals(mack.model)
  resid$dev.in.units <- sort(unique(triangle.long$dev))[resid$dev.period]
  loss.plot <- (qplot(x=fitted.value, y=standard.residuals, data=resid)
      + geom_hline(aes(x=0), color=favir.colors["M2"])
      + geom_smooth(aes(x=fitted.value, y=standard.residuals), se=FALSE)
      + labs(x="Predicted Loss", y="Standardized Residual"))
  origin.plot <- (ggplot(data=resid)
      + geom_point(aes(x=origin.period, y=standard.residuals))
      + geom_hline(aes(x=0), color=favir.colors["M2"])
      + geom_smooth(aes(x=origin.period, y=standard.residuals), se=FALSE)
      + labs(x=origin.field.heading, y="Standardized Residual")
      + xlim(range(resid$origin.period)))
  cal.plot <- (qplot(x=cal.period, y=standard.residuals, data=resid)
      + geom_hline(aes(x=0), color=favir.colors["M2"])
      + geom_smooth(aes(x=cal.period, y=standard.residuals), se=FALSE)
      + labs(x="Calendar Year", y="Standardized Residual")
      + xlim(range(resid$origin.period)))
  dev.plot <- (qplot(x=dev.in.units, y=standard.residuals, data=resid)
      + geom_hline(aes(x=0), color=favir.colors["M2"])
      + geom_smooth(aes(x=dev.in.units, y=standard.residuals), se=FALSE)
      + labs(x="Starting Development Age", y="Standardized Residual")
      + scale_x_continuous(breaks=resid$dev.in.units))
  return(list(loss=loss.plot, origin=origin.plot, cal=cal.plot, dev=dev.plot))
}

RegressionModels <- function(triangle.df, premium.df) {
  # Return a list of lists of alternate age-to-age models
  #
  # Args:
  # triangle.df - a triangle in long data-frame form
  # premium.df - the premium by year
  #
  # Output:
  # A list with length the number of age-to-age development links.
  # Each element in the list corresponds to the output of MakeDevModels
  # Make a triangle of loss ratios, to adjust for changing premium amounts
  triangle.df$value <- (triangle.df$value /
           LookupDFMult(premium.df, origin=triangle.df$origin)$premium)
  loss.ratio.tri <- as.triangle(triangle.df)

  mack.lr.model <- ChainLadder::MackChainLadder(loss.ratio.tri)
  Assert(length(mack.lr.model$Models) == ncol(loss.ratio.tri) - 1,
         "Expected one model for each triangle column, except last")
  model.names <- paste(head(colnames(loss.ratio.tri), -1), "to",
                       tail(colnames(loss.ratio.tri), -1))
  alt.dev.models <- lapply(Indicies(mack.lr.model$Models),
                           function(i) MakeDevModels(model.names[i],
                                                     mack.lr.model$Models[[i]]))
  names(alt.dev.models) <- model.names
  return(alt.dev.models)
}

RegressionGraph <- function(alt.dev.models) {
  # Return graph of regression results given list of models
  #
  # Args: list of models as produced by RegressionModels
  #
  # Output: Graph summarizing the regression
  summary.regression.df <- do.call(rbind,
               lapply(alt.dev.models,
                      function(model.list) do.call(GetLinkDF, model.list)))
  color_manual_values <- c(favir.colors["A5"], favir.colors["M3"],
                           favir.colors["M4"])
  names(color_manual_values) <- c("Link Only", "Intercept Only", "Both Terms")
  reg.graph <- (ggplot(data=summary.regression.df) + facet_wrap(~ dev)
                + geom_line(aes(x=start, y=link.only, colour="Link Only"))
                + geom_line(aes(x=start, y=both, colour="Both Terms"))
                + geom_line(aes(x=start, y=intercept.only,
                                colour="Intercept Only"))
                + scale_colour_manual("Model", color_manual_values)
                + geom_point(aes(x=start, y=actual)))
  ylim <- range(summary.regression.df$actual
                [!is.na(summary.regression.df$actual)])
  yrange <- ylim[2] - ylim[1]
  reg.graph <- (reg.graph + labs(x="Starting Loss Ratio",
                                 y="Increase in Loss Ratio")
                + coord_cartesian(ylim=c(ylim[1] - 0.05 * yrange,
                                         ylim[2] + 0.05 * yrange))
                + scale_y_continuous(breaks=round(ggplot2::breaks(ylim,
                                       "width", 4), 2)))
  return(reg.graph)
}

RegressionStats <- function(alt.dev.models) {
  # Return favir data frame of regression statistics
  #
  # Args: list of models as produced by RegressionModels
  #
  # Output: a favir data frame summarizing regression statistics
  reg.stats.fdf <- TransposeDF(lapply(alt.dev.models, DevModelStats))
  reg.stats.fdf <- Favir.data.frame(reg.stats.fdf)
  FieldHeadings(reg.stats.fdf) <- list(dev.period="Development Period",
                                       link.RS="Link Only $R^2$ \\%",
                                       both.RS="$R^2$ \\%",
                                       intercept.t="Intercept $t$-Value",
                                       link.t="Link $t$-Value",
                                       intercept.power="Intercept $p$-val \\%",
                                       link.power="Link Power $p$-val \\%")
  FieldGroup(reg.stats.fdf, "both") <- 3:7
  GroupHeading(reg.stats.fdf, "both") <- "Fit Results: Link and Intercept Model"
  FieldGroup(reg.stats.fdf, "percent") <- c("link.RS", "both.RS",
                                            "intercept.power", "link.power")
  return(reg.stats.fdf)
}


################################################################
# latex.R - code for making pretty tables

######### Color definitions and graphics defaults
# Three sets of colors in color scheme: labeled M (color main),
# A (alternate A), and B (alternateB).  In each group, the
# five colors are about the same hue and arranged from lightest to darkest
favir.colors <- c(M1="#F1EBE6", # Main color defaults to brown
                  M2="#D3BDA8", M3="#B8987C", M4="#7C6C5C", M5="#583008",
                  A1="#C3E5D0", # Alternate 1 is a green
                  A2="#A4E5BC", A3="#569C70", A4="#435A4B", A5="#06401B",
                  B1="#C0CDE9", # Alternate 2 is a blue
                  B2="#7C93C5", B3="#435D94", B4="#35425D", B5="#0A1F4B")

Favir.data.frame <- function(df, field.formatters=list(),
                             field.headings=list(), default.formatter=NULL,
                             row.colors=NULL, orientation="normal",
                             caption="", label=NULL, text.size="normalsize") {
  # Return version of data frame df with latex annotation
  #
  # Args:
  # df - a data frame
  #
  # Output: An object of type favir.data.frame, which is like
  # the original object but contains formatting suggestions.
  class(df) <- unique(c("favir.data.frame", class(df)))
  len <- length(df)

  # Set default formatter
  if (!is.null(default.formatter))
    DefaultFormatter(df) <- default.formatter

  # Set field formatters
  attributes(df)$favir.field.formatters <- list()
  FieldFormatters(df) <- field.formatters

  # Set field headings
  attributes(df)$favir.field.headings <- list()
  FieldHeadings(df) <- field.headings

  # Set row colors
  if (is.null(row.colors))
    RowColors(df) <- DefaultColoring(df)
  else RowColors(df) <- row.colors

  TextSize(df) <- text.size

  # Set field groups and group headings
  attributes(df)$favir.field.groups <- list()
  attributes(df)$favir.group.headings <- list()
  attributes(df)$favir.orientation <- orientation
  attributes(df)$favir.caption <- caption
  attributes(df)$favir.label <- label
  attributes(df)$favir.summary.row <- NULL
  
  return(df)
}

DefaultFormatter <- function(fdf) {
  # Return the default cell formatting function
  return(attributes(fdf)$favir.default.formatter)
}

`DefaultFormatter<-` <- function(fdf, value) {
  # Set the default formatting function function
  Assert(mode(value) == "function", "Formatter must be a function")
  attributes(fdf)$favir.default.formatter <- value
  return(fdf)
}

MakeFormatter <- function(digits=2, format="f", big.mark=",",
                          omit.math=FALSE, ...) {
  # Short cut for making a formatter; parameters passed to formatC
  #
  # Args: list of arguments acceptable to the base package's formatC function
  #
  # Output: a function from number vectors to character vectors
  params <- list(digits=digits, format=format, big.mark=big.mark, ...)
  Formatter <- function(val) {
    # This is the actual formatting function
    formatc.input <- ifelse(is.na(val) | is.null(val), 0, val)
    formatc.output <- do.call(formatC, c(list(x=formatc.input), params))
    Helper <- function(single) {
      # Convert a single value from number to character
      if (omit.math)
        return(single)
      return(paste("$", single, "$", sep=""))
    }
    return(sapply(ifelse(is.na(val) | is.null(val), "", formatc.output), Helper))
  }
  return(Formatter)
}

.globals <- list( # This holds global formatting options
  default.formatter=MakeFormatter(), # Default global formatter
  # The list below maps field group names to formatting functions
  field.group.formatters=list()
)

SetGlobalDefaultFormatter <- function(formatter) {
  # Warning, sets global default formatter to given function
  .globals$default.formatter <<- formatter
}

FieldFormatters <- function(fdf) {
  # Return list from df headers to cell formatting functions
  return(attributes(fdf)$favir.field.formatters)
}

`FieldFormatters<-` <- function(fdf, value) {
  # Set the column formatting functions
  Assert(all(names(value) %in% names(fdf)), "Formatted specified for bad col")
  for (name in names(value))
    attributes(fdf)$favir.field.formatters[name] <- value[name]
  return(fdf)
}

FieldHeadings <- function(fdf) {
  # Return list that maps df names to their printable field titles
  return(attributes(fdf)$favir.field.headings)
}

`FieldHeadings<-` <- function(fdf, value) {
  # Set the column formatting functions
  Assert(all(names(value) %in% names(fdf)), "Heading specified for bad col")
  for (name in names(value))
    attributes(fdf)$favir.field.headings[name] <- value[name]
  return(fdf)
}

RowColors <- function(fdf) {
  # Return the row coloring vector
  #
  # Args: fdf - the favir data frame
  #
  # Output: a character vector, one for each row, of latex colors
  return(attributes(fdf)$favir.row.colors)
}

`RowColors<-` <- function(fdf, value) {
  # Set the row colors
  #
  # Args: fdf - the favir data frame to update
  # value - a vector of character color names (one per row of data frame)
  #
  # Output: This updates 
  Assert(length(value) == nrow(fdf), "Color vec doesn't match number of rows")
  attributes(fdf)$favir.row.colors <- value
  return(fdf)
}

DefaultColoring <- function(df, primary="white", secondary="colorM1") {
  # Return the default row coloring for data frame df
  if (nrow(df) <= 5)
    return(rep(primary, nrow(df)))
  return(ifelse((1:nrow(df) - 1) %% 6 <= 2, primary, secondary))
}

FieldSettings <- function(fdf, field) {
  # Return field settings for the given df field
  #
  # Args:
  # fdf - a favir.data.frame
  # field - name of the fields in the data frame
  #
  # Output: a list with the following elements
  # heading - the printable field title
  # formatter - a function used to format the data in that field
  Assert(length(field) == 1 && field %in% names(fdf), "Bad column definition")
  result <- list()

  field.headings <- FieldHeadings(fdf)
  if (field %in% names(field.headings))
    result$heading <- field.headings[[field]]

  formatters <- FieldFormatters(fdf)
  if (field %in% names(formatters))
    result$formatter <- formatters[[field]]

  return(result)
}

`FieldSettings<-` <- function(fdf, field, value) {
  # Update the field settings for given field
  #
  # Args:
  # fdf - a favir.data.frame
  # field - the field name to update
  # value - a list of field settings, with names field.heading and
  #         field.formatter
  #
  # Output: the result is updating the fdf
  Assert(length(field) == 1 && field %in% names(fdf), "Bad column name given")
  Assert(all(names(value) %in% c("heading", "formatter")), "Unknown setting")
  if ("heading" %in% names(value))
    FieldHeadings(fdf)[[field]] <- value$heading
  if ("formatter" %in% names(value)) {
    Assert(mode(value$formatter) == "function", "Formatter must be a function")
    FieldFormatters(fdf)[[field]] <- value$formatter
  }
  return(fdf)
}

ListFieldGroups <- function(fdf) {
  # Return a character vector contining all existing field groups names
  return(names(attributes(fdf)$favir.field.groups))
}

FieldGroup <- function(fdf, group.name, as.names=TRUE) {
  # Return the existing fields in the given group, or NULL if no group exists
  Assert(group.name %in% ListFieldGroups(fdf), "Unknown Group")
  field.indicies <- attributes(fdf)$favir.field.groups[[group.name]]
  if (as.names)
    return(names(fdf)[field.indicies])
  else return(field.indicies)
}

`FieldGroup<-` <- function(fdf, group.name, value) {
  # Set the field group with given name to contain the fields given in value
  Assert(length(unique(value)) == length(value), "Some field repeated")
  if (mode(value) == "character") { # value is vec of field names
    Assert(all(value %in% names(fdf)), "Bad field given")
    value <- which(names(fdf) %in% value)
  } else if (mode(value) == "numeric") { # value is vec of field indicies
    Assert(all(value %in% Indicies(fdf)), "Numeric field out of range")
  } else stop("Invalid type for field group assignment")
  attributes(fdf)$favir.field.groups[[group.name]] <- value
  return(fdf)
}

ParentFieldGroups <- function(fdf, field) {
  # Return a vector of field group names that contain the given field
  #
  # field is interpreted as a name or number depending on its class
  AssertSingle(field)
  if ("character" %in% class(field))
    predicate <- function(group) field %in% FieldGroup(fdf, group)
  else predicate <- function(group) field %in% FieldGroup(fdf, group,
                                                          as.names=FALSE)
  return(Filter(predicate, ListFieldGroups(fdf)))
}

FieldGroupFormatter <- function(fdf, group.name) {
  # Return the group formatter for group.name, or NULL if there isn't one
  return(attributes(fdf)$favir.group.formatters[[group.name]])
}

`FieldGroupFormatter<-` <- function(fdf, group.name, value) {
  # Set the field group formatter for group.name to formatter
  AssertSingle(group.name, value)
  Assert("function" %in% class(value), "Formatter must be a function")
  attributes(fdf)$favir.group.formatters[[group.name]] <- value
  return(fdf)
}

GetGlobalFieldGroupFormatter <- function(group.name) {
  # Return the global field formatter for the given group, or NULL
  return(.globals$field.group.formatters[[group.name]])
}

SetGlobalFieldGroupFormatter <- function(group.name, formatter) {
  # Set globally the given formatter to be used for that field group
  #
  # Args:
  # group.name - the name of a field group
  # formatter - the cell formatting function to use for that group's data
  #
  # 
  AssertSingle(group.name, formatter)
  Assert("function" %in% class(formatter), "Second arg needs to be function")
  .globals$field.group.formatters[[group.name]] <<- formatter
}

# Set a couple of common defaults
SetGlobalFieldGroupFormatter("money", MakeFormatter(digits=0))
SetGlobalFieldGroupFormatter("origin", MakeFormatter(digits=0, big.mark=""))
SetGlobalFieldGroupFormatter("LDF", MakeFormatter(digits=2, big.mark=""))
SetGlobalFieldGroupFormatter("dev", MakeFormatter(digits=0, big.mark=""))
PercentFormatter <- function(x) MakeFormatter(digits=1, big.mark="")(x*100)
SetGlobalFieldGroupFormatter("percent", PercentFormatter)

IsGroupContiguous <- function(fdf, group.name) {
  # Returns TRUE if the group joins contiguous fields, false otherwise
  field.indicies <- FieldGroup(fdf, group.name, as.names=FALSE)
  highlo <- range(field.indicies)
  if (length(field.indicies) - 1 == highlo[2] - highlo[1])
    return(TRUE)
  return(FALSE)
}

GroupHeading <- function(fdf, group.name) {
  # Return the heading for given group, or NULL if group has no heading
  Assert(length(group.name) == 1 && group.name %in% ListFieldGroups(fdf))
  return(attributes(fdf)$favir.group.headings[[group.name]])
}

`GroupHeading<-` <- function(fdf, group.name, value) {
  # Set the displayed heading for group to the given value
  Assert(length(group.name) == 1 && group.name %in% ListFieldGroups(fdf))
  AssertSingle(length(value) == 1, "Group heading should be a single string")
  Assert(IsGroupContiguous(fdf, group.name),
         "Group headings can only be applied to contiguous fields")
  attributes(fdf)$favir.group.headings[[group.name]] <- value
  return(fdf)
}

SortedGroupHeadings <- function(fdf) {
  # Return char vec of groups with headings in column order
  GetFirstCol <- function(group.name) {
    # Return the index of the first column in the given group
    return(min(FieldGroup(fdf, group.name, as.names=FALSE)))
  }

  groups <- names(attributes(fdf)$favir.group.headings)
  if (length(groups) == 0)
    return(NULL)
  first.cols <- sapply(groups, GetFirstCol)
  return(groups[order(first.cols)])
}

Orientation <- function(fdf) {
  # Return the orientation of the table
  return(attributes(fdf)$favir.orientation)
}

`Orientation<-` <- function(fdf, value) {
  # Set the orientation for the table.
  Assert(identical(value, "normal") || identical(value, "sideways"))
  attributes(fdf)$favir.orientation <- value
  return(fdf)
}

Caption <- function(fdf) {
  # Return the caption of the table
  return(attributes(fdf)$favir.caption)
}

`Caption<-` <- function(fdf, value) {
  # Set the caption string for the table.
  AssertSingle(value, "Caption should be a single string")
  attributes(fdf)$favir.caption <- value
  return(fdf)
}

Label <- function(fdf) {
  # Return the table's latex reference label
  return(attributes(fdf)$favir.label)
}

`Label<-` <- function(fdf, value) {
  # Set the table's latex reference label
  AssertSingle(value, "Label should be a single string")
  attributes(fdf)$favir.label <- value
  return(fdf)
}

SummaryRow <- function(fdf) {
  # Return the table's summary row
  #
  # The summary row is a row below the table set off by an hline that
  # can hold totals or averages.
  #
  # Args: fdf - favir data frame to process
  #
  # Result: A list mapping data frame fields to summary row contents, or NULL
  return(attributes(fdf)$favir.summary.row)
}

`SummaryRow<-` <- function(fdf, value) {
  # Set the table's summary row.  See SummaryRow() comments.
  Assert(is.null(value) || "list" %in% class(value), "Expecting list as arg")
  if (!is.null(value))
    Assert(all(names(value) %in% names(fdf)), "Invalid field name")
  attributes(fdf)$favir.summary.row <- value
  return(fdf)
}

TextSize <- function(fdf) {
  # Get the favir data frame's text size
  return(attributes(fdf)$favir.text.size)
}

`TextSize<-` <- function(fdf, value) {
  # Set the fdf's text size
  AssertSingle(value)
  options <- c("tiny", "scriptsize", "footnotesize", "small", "normalsize",
               "large", "Large", "LARGE", "huge", "Huge")
  Assert(value %in% options, "Invalid text size given")
  attributes(fdf)$favir.text.size <- value
  return(fdf)
}

MatrixToFDF <- function(matrix, ..., rowname.col=NULL) {
  # Convenience function: convert matrix in one step to favir data frame
  #
  # if row.name.col is not NULL, row names are added to the data frame
  # under a separate name
  df <- as.data.frame(matrix)
  if (!is.null(rowname.col)) {
    rowname.df <- data.frame(rownames(matrix))
    names(rowname.df) <- rowname.col
    df <- cbind(rowname.df, df)
  }
  return(Favir.data.frame(df, ...))
}

            
######### Latex code detail section - create latex from a favir df

SplitHeadings <- function(char.vec, threshold=11, max.rows=3,
                          split.char=" ") {
  # Split the character vector into multiple rows
  #
  # Input:
  # char.vec - a vector of strings to split
  # threshold - the number of characters per string required to split
  # max.rows - the maximum number of rows to split the string into
  #
  # Output: a matrix of strings.  Each column corresponds to the same
  # entry in char.vec.  Unused entries, from the top, will be filled
  # with the empty string "".
  SplitSingle <- function(s) {
    # Split a single string into a vector of multiple strings
    str.comps <- StrSplit(s, split.char)
    split.list <- SplitHelper(str.comps, max.rows)
    return(sapply(split.list, Collapse))
  }

  SplitHelper <- function(char.vec, max.rows) {
    # Return a list of string components
    #
    # The basic idea is to start from the end, and keep prepending
    # words until we are over the threshold, then move on to the
    # previous line.
    if (max.rows == 1 || length(char.vec) <= 1)
      return(list(char.vec))
    cur <- Last(char.vec)
    for (i in (length(char.vec) - 1):1) {
      if (nchar(Collapse(c(char.vec[i], cur))) > threshold) {
        split.rest <- SplitHelper(char.vec[1:i], max.rows - 1)
        return(c(split.rest, list(cur)))
      }
      cur <- c(char.vec[i], cur)
    }
    return(list(cur))
  }

  Collapse <- function(char.vec) {
    # Join the pieces of char.vec together again and return single string
    return(paste(char.vec, collapse=split.char))
  }
  
  result.list <- lapply(char.vec, SplitSingle)
  num.rows <- max(sapply(result.list, length))
  result <- matrix("", nrow=num.rows, ncol=length(char.vec))
  for (i in Indicies(char.vec)) {
    splat <- result.list[[i]]
    result[(num.rows - length(splat) + 1):num.rows, i] <- splat
  }
  return(result)
}

kTablePrefix1 <- "\\begin{figure}"
kTablePrefix2 <- "
\\begin{center}
\\begin{tikzpicture}
\\node (tbl) {
\\begin{tabular}"
kTableSuffix1 <- "\\end{tabular}};
\\begin{pgfonlayer}{background}
\\draw[rounded corners,left color=colorM2,right color=colorM3,draw=colorM5]
    ($(tbl.north west)+(0.14,0.00)$) rectangle
           ($(tbl.south east)-(0.14,0.00)$);
\\end{pgfonlayer}
\\end{tikzpicture}"
kTableSuffix2 <- "\\end{center}\\end{figure}\n"
kSidewaysTablePrefix1 <- "\\begin{sidewaysfigure}"
kSidewaysTableSuffix2 <- "\\end{center}\\end{sidewaysfigure}\n"

ColumnTypes <- function(data.matrix) {
  # Return the latex column types for each column
  # Below, subtract 2 because of the math mode symbols $
  data.lengths <- apply(data.matrix, 2,
                        function(col) max(nchar(col)) - 2)
  return(paste("D{!}{!}{", data.lengths, ".0}", sep=""))
}

PreLatex <- function(favir.df) {
  # Return a pre-latex version favir data frame 
  #
  # Args:
  # favir.df - an object of class favir.data.frame
  #
  # Output: An object of type favir.prelatex.table, which can easily
  # be transformed into latex
  GetFormatter <- function(field.index) {
    # Return the formatting function for the given field
    #
    # Here is the order of precedence given to conflicting field
    # formatting directives:
    #   1.  Explicit local (to favir.df) field formatter
    #   2.  Local field group formatter (in alphabetical order)
    #   3.  Global field group formatter (in alphabetical order)
    #   4.  Local default formatter
    #   5.  Global default formatter
    AssertSingle(field.index)
    field.name <- names(favir.df)[field.index]
    ff <- FieldFormatters(favir.df)[[field.name]]
    if (!is.null(ff)) # Choice 1
      return(ff)

    field.groups <- ParentFieldGroups(favir.df, field.index)
    for (field.group in field.groups)
      if (!is.null(FieldGroupFormatter(favir.df, field.group)))
        return(FieldGroupFormatter(favir.df, field.group)) # Choice 2        
    for (field.group in field.groups)
      if (!is.null(GetGlobalFieldGroupFormatter(field.group)))
        return(GetGlobalFieldGroupFormatter(field.group)) # Choice 3

    if (!is.null(DefaultFormatter(favir.df)))
      return(DefaultFormatter(favir.df)) # Choice 4
    return(.globals$default.formatter) # Choice 5
  }

  DataMatrix <- function() {
    # Return character matrix of formatted values
    m <- matrix(NA, nrow=nrow(favir.df), ncol=length(favir.df))
    for (i in Indicies(favir.df)) {
      data.col <- favir.df[[i]]
      if ("factor" %in% class(data.col))
        data.col <- as.character(data.col)
      m[, i] <- GetFormatter(i)(data.col)
    }
    return(m)
  }

  DataRowPrefixes <- function(row.colors) {
    # Return prefixes to each data row
    return(paste("\\rowcolor{", row.colors, "}", sep=""))
  }

  GroupRow <- function() {
    # Return string containing the latex code for the group headings
    groups <- SortedGroupHeadings(favir.df)
    if (length(groups) == 0)
      return(NULL)

    heading.vec <- NULL
    last.col <- 0
    for (group in groups) {
      begin.col <- min(FieldGroup(favir.df, group, as.names=FALSE))
      heading.vec <- c(heading.vec, rep("", begin.col - last.col - 1))
      last.col <- max(FieldGroup(favir.df, group, as.names=FALSE))
      col.length <- last.col - begin.col + 1
      heading <- GroupHeading(favir.df, group)
      heading.vec <- c(heading.vec, paste("\\multicolumn{", col.length,
                                          "}{c}{", heading, "}", sep=""))
    }
    return(paste(heading.vec, collapse=" & "))
  }

  GroupRules <- function() {
    # Return the latex code to make midrules after the group headings
    groups <- SortedGroupHeadings(favir.df)
    if (length(groups) == 0)
      return(NULL)
    begin.cols <- sapply(groups,
              function(g) min(FieldGroup(favir.df, g, as.names=FALSE)))
    end.cols <- sapply(groups,
              function(g) max(FieldGroup(favir.df, g, as.names=FALSE)))
    return(paste("\\cmidrule(rl){", begin.cols, "-", end.cols, "}",
                 sep="", collapse=""))
  }
  
  HeadingMatrix <- function() {
    # Return heading matrix with long headings split between rows
    headings <- names(favir.df)
    for (i in Indicies(headings))
      if (headings[i] %in% names(FieldHeadings(favir.df)))
        headings[i] <- FieldHeadings(favir.df)[[headings[i]]]
    return(SplitHeadings(headings))
  }

  FormattedHeadings <- function(heading.matrix) {
    # Return heading matrix enclosed in a multicolumn environment
    result <- paste("\\multicolumn{1}{c}{", heading.matrix, "}", sep="")
    dim(result) <- dim(heading.matrix)
    return(result)
  }

  PrefixAndSuffix <- function() {
    # Return a list that has prefix, suffix1 and suffix 2 set
    result <- list()
    if (identical(Orientation(favir.df), "normal")) {
      # Table has normal orientation
      result$prefix1 <- kTablePrefix1
      result$suffix2 <- kTableSuffix2
    } else { # Table is sideways
      result$prefix1 <- kSidewaysTablePrefix1
      result$suffix2 <- kSidewaysTableSuffix2
    }
    result$prefix2 <- kTablePrefix2
    result$suffix1 <- kTableSuffix1
    return(result)
  }

  FormatSummaryRow <- function() {
    # Return the formatted entries in the summary row as a vector, or NULL
    if (is.null(SummaryRow(favir.df)))
      return(NULL)

    Helper <- function(field.index) {
      # Return formatted row entry
      field <- names(favir.df)[field.index]
      raw.entry <- SummaryRow(favir.df)[[field]]
      if (is.null(raw.entry))
        return("")
      #if ("character" %in% class(raw.entry)) # center character strings
      #  return(paste("\\multicolumn{1}{c}{", raw.entry, "}", sep=""))
      return(GetFormatter(field.index)(raw.entry))
    }

    formatted <- sapply(Indicies(favir.df), Helper)
    hline <- "\\arrayrulecolor{colorM5}\\hline"
    # use first row's color
    row.color <- DataRowPrefixes(RowColors(favir.df)[1]) 
    strut <- "\\rule{0pt}{1.2em}"
    formatted[1] <- paste(hline, row.color, strut, formatted[1], sep="")
    return(formatted)
  }
  
  result <- PrefixAndSuffix()
  result$text.size <- paste("\\", TextSize(favir.df), sep="")
  result$group.row <- GroupRow()
  result$group.rules <- GroupRules()
  result$heading.matrix <- HeadingMatrix()
  result$formatted.headings <- FormattedHeadings(result$heading.matrix)
  result$data.prefix <- "\\addlinespace"
  result$data.matrix <- DataMatrix()
  # extra data to be inserted before any data row
  result$extra.rows <- rep("", nrow(result$data.matrix))
  result$column.types <- ColumnTypes(result$data.matrix)
  result$caption <- paste("\\caption{", Caption(favir.df), "}",
                          collapse="", sep="")
  result$label <- paste("\\label{", Label(favir.df), "}",
                        collapse="", sep="")
  result$heading.row.prefixes <- rep("", nrow(result$heading.matrix))
  result$data.row.prefixes <- DataRowPrefixes(RowColors(favir.df))
  result$summary.row <- FormatSummaryRow()
  
  class(result) <- "favir.prelatex"
  return(result)
}

print.favir.prelatex <- function(pl) {
  # Print a PreLatex object
  Assert("favir.prelatex" %in% class(pl))
  column.spec <- paste("{", paste(pl$column.types, collapse=""), "}",
                       collapse="")

  if (is.null(pl$group.row))
    group.code <- ""
  else group.code <- paste(pl$group.row, "\\\\\n", pl$group.rules,
                           "\n", collapse="")

  heading.rows <- apply(pl$formatted.headings, 1,
                        function(...) paste(..., collapse=" & "))
  heading <- paste(pl$heading.row.prefixes, heading.rows,
                   sep="\n", collapse="\\\\\n")
  
  data.rows <- apply(pl$data.matrix, 1,
                     function(...) paste(..., collapse=" & "))
  data <- paste(pl$extra.rows, pl$data.row.prefixes,
                data.rows, sep="\n", collapse="\\\\\n")
  summary.row <- ifelse(is.null(pl$summary.row), "",
                        paste("\\\\\n", paste(pl$summary.row, collapse=" & ")))
  result <- paste(pl$prefix1, pl$text.size, pl$prefix2, column.spec, group.code,
                  heading, "\\\\\n", pl$data.prefix, data, summary.row,
                  pl$suffix1, pl$caption, pl$label, pl$suffix2, collapse="")
  cat(result)
  return(invisible(result))
}

RBindPreLatex <- function(..., include.headings=TRUE) {
  # Join number of PreLatex objects together in a single PreLatex object
  RBindLenient <- function(matrix1, matrix2) {
    # Bind matricies 1 and 2 together on row, but allow for column padding
    Pad <- function(m, n) {
      # Add columns of "" to matrix m so it's number of col is n
      padding <- matrix("", nrow=nrow(m), ncol=n - ncol(m))
      return(cbind(m, padding))
    }
    if (ncol(matrix1) == ncol(matrix2))
      return(rbind(matrix1, matrix2))
    if (ncol(matrix1) < ncol(matrix2))
      return(rbind(Pad(matrix1, ncol(matrix2)), matrix2))
    return(rbind(matrix1, Pad(matrix2, ncol(matrix1))))
  }

  args <- list(...)
  Assert(all(sapply(args, function(obj) "favir.prelatex" %in% class(obj))),
         "RBindPreLatex takes prelatex objects as arguments")
  Assert(length(args) >= 2)
  result <- args[[1]]
  for (i in 2:length(args)) {
    if (include.headings) {
      if (!is.null(args[[i]]$group.row))
        group.code <- paste(args[[i]]$group.row, "\\\\\n", args[[i]]$group.rules,
                            "\n", collapse="")
      else group.code <- ""
      heading.rows <- apply(args[[i]]$formatted.headings, 1,
                            function(...) paste(..., collapse=" & "))
      heading.code <- paste(args[[i]]$heading.row.prefixes, heading.rows,
                            sep="\n", collapse="\\\\\n")
      extra.row <- paste("\\addlinespace", group.code, heading.code, "\\\\",
                         args[[i]]$data.prefix, args[[i]]$extra.rows[1],
                         sep="", collapse="")
    } else {
      #num.col <- max(ncol(result$data.matrix), ncol(args[[i]]$data.matrix))
      #rule <- paste("\\cmidrule(rl){1-", num.col, "}\n", sep="")
      extra.row <- paste(args[[i]]$data.prefix, args[[i]]$extra.rows[1],
                         sep="", collapse="")
    }
    result$extra.rows <- c(result$extra.rows, extra.row,
                           tail(args[[i]]$extra.rows, -1))
    result$data.matrix <- RBindLenient(result$data.matrix,
                                       args[[i]]$data.matrix)
    result$data.row.prefixes <- c(result$data.row.prefixes,
                                  args[[i]]$data.row.prefixes)
    result$column.types <- ColumnTypes(result$data.matrix)
  }
  return(result)
}

print.favir.data.frame <- function(fdf) {
  # Print a favir data frame
  Assert("favir.data.frame" %in% class(fdf))
  return(invisible(print(PreLatex(fdf))))
}

#######################################################################
# Non-table related latex code

IncludeGraph <- function(graph, ...) {
  # Write the graph to a PDF file and produce the latex to include the graph
  #
  # Arguments:
  # graph - the ggplot object to include
  # filename - file to save plot to.  Otherwise derive name from graph
  # caption - the latex caption for the graph
  # label - the latex reference label for the graph
  # width, height - dimensions in inches for the graph
  #
  # Output: No return value, but create a PDF file on disk and print
  # latex code necessary to include PDF in a document.
  thunk <- function() print(graph)
  graph.name <- as.character(as.expression(as.list(match.call())[[2]]))
  IncludeGraphHelper(thunk, graph.name, ...)
}

IncludeGrid <- function(graph.list, ...) {
  # Write a grid of graphs to PDF and output latex
  #
  # Arguments: like IncludeGraph, but take graph.list instead of graph
  # graph.list - a list of graphs or other printable objects.  The keys must
  #              have the form "x.y" to indicate the row and column number.
  ParseName <- function(key) {
    # Parse the graph list names and return two vectors
    split <- StrSplit(key, ".")
    Assert(length(split) == 2, "List names must have the form 'XXX.YYY'")
    coords <- as.integer(split)
    Assert(all(coords > 0 && coords < 99), "Bad coordinates given")
    return(coords)
  }

  Assert(length(graph.list) == length(names(graph.list)), "Unnamed element")
  Assert(length(graph.list) >= 2, "Call IncludeGrid with two or more graphs")
  args <- list(...)
  if ("label" %in% names(args))
    graph.name <- args$label
  else {
    Assert("caption" %in% names(args),
              "Unable to find name for your graph: provide label or caption")
    graph.name <- args$caption
  }

  coord.matrix <- sapply(names(graph.list), ParseName)
  Assert(nrow(coord.matrix) == 2, "Expected a two dimensional grid")
  num.viewports <- c(max(coord.matrix[1, ]), max(coord.matrix[2, ]))
    
  thunk <- function() {
    # IncludeGraphHelper runs this to actually print the total graph
    grid.newpage()
    pushViewport(viewport(layout=grid.layout(num.viewports[1],
                                             num.viewports[2])))
    for (i in 1:length(graph.list))
      print(graph.list[[i]], vp=viewport(layout.pos.row=coord.matrix[1, i],
                                         layout.pos.col=coord.matrix[2, i]))
  }

  IncludeGraphHelper(thunk, graph.name, ...)
}

IncludeGraphHelper <- function(thunk, graph.name, filename=NULL, caption="",
                               label=NULL, width=5, height=5) {
  # Helper function for IncludeGraph and IncludeGrid
  AssertSingle(caption, width, height)
  if (is.null(filename)) { # Extract filename candidate from graph name
    # Now get rid of spaces and periods, which latex doesn't like in filenames
    prefix <- gsub(" ", "_", gsub(".", "_", graph.name, fixed=TRUE), fixed=TRUE)
    filename <- paste("graphs/", prefix, ".pdf", sep="")
  }
  pdf(file=filename, paper="special", width=width, height=height)
  thunk()
  dev.off()
  cat(paste("\\begin{figure}\\begin{center}\\includegraphics[width=",
            width, "in,height=", height, "in]{",
            filename, "}\\caption{", caption, "}", sep=""))
  if (!is.null(label))
    cat(paste("\\label{", label, "}", sep=""))
  cat("\\end{center}\\end{figure}")
}

MakeColorLines <- function(named.color.vec) {
  # Return latex character string defining the given colors
  #
  # Argument is a vector of colors like "#35425D".  Each should be
  # named; each name will be defined to be that color
  #
  # Output: color definition string lines that look like this:
  # \definecolor{colorM1}{RGB}{242,236,231}
  # where M1 was the name of the color in this case
  Helper <- function(color) {
    # Return a color triple like "88,248,8"
    return(paste(as.vector(col2rgb(color)), collapse=","))
  }
  stopifnot(length(names(named.color.vec)) == length(named.color.vec))
  return(paste("\\definecolor{color", names(named.color.vec), "}{RGB}{",
               sapply(named.color.vec, Helper), "}\n", sep=""))
}

ThemePlot <- function() {
  # Set the theming parameters for ggplot2
  theme_update(panel.background=theme_rect(fill=favir.colors["M1"],
                                           colour=favir.colors["M1"]),
               legend.key=theme_rect(fill=favir.colors["M1"],
                                     colour=favir.colors["M1"]),
               strip.background=theme_rect(fill=favir.colors["M2"],
                                           colour=favir.colors["M2"]))
  update_geom_defaults("point", aes(colour=favir.colors["M5"],
                                    size=2.5, shape=1))
  update_geom_defaults("jitter", aes(colour=favir.colors["M5"],
                                     size=2.5, shape=1))
  update_geom_defaults("line", aes(colour=favir.colors["A5"]))
  update_geom_defaults("abline", aes(colour=favir.colors["A5"]))
  update_geom_defaults("smooth", aes(colour=favir.colors["A5"]))
  update_geom_defaults("bar", aes(colour=favir.colors["M3"],
                                  fill=favir.colors["M3"]))
}
ThemePlot()

InsertPrelude <- function() {
  # Insert the latex prelude into the document
  cat("\\documentclass[letterpaper,12pt]{book}
\\usepackage[hmargin=1in,bmargin=1.5in,tmargin=1.2in]{geometry}
\\usepackage{color}
\\usepackage{booktabs}
\\usepackage{colortbl}
\\usepackage{dcolumn}
\\usepackage{multirow}
\\usepackage{rotating}
\\usepackage{tabularx}

\\usepackage{tikz}
\\usetikzlibrary{calc}
\\pgfdeclarelayer{background}
\\pgfdeclarelayer{foreground}
\\pgfsetlayers{background,main,foreground}")
  cat(MakeColorLines(favir.colors))
  cat("%%%%%%%%%%%%%%% End Colors, Start Logo

\\newcommand{\\favirlogo}{
\\begin{minipage}{3cm}  
\\fontfamily{pcr}\\bfseries\\Huge
\\begin{tikzpicture}
  % Length and size definitions
  \\def\\logobackwidth{1.3}
  \\def\\logobackheight{0.5}
  \\def\\logobackradius{0.15}
  \\def\\logobackgap{0.03}
  \\def\\logotextjitter{0.05}

  % Fill in one background rectangle by tracing clockwise path
  \\def\\logopath{(\\logobackgap, \\logobackgap) -- +(0, \\logobackheight)
                         --  +($(\\logobackwidth, \\logobackheight)
                                    -(\\logobackradius, 0)$)
                         arc (90:0:\\logobackradius)
                         --  ++($(0, \\logobackradius)-(0, \\logobackheight)$)
                         --  +(\\logobackwidth,0)
                         -- cycle}
  \\fill[colorM1] \\logopath;
  % Then copy it with two reflections and one rotation
  \\fill[colorM1,cm={1,0,0,-1,(0,0)}] \\logopath;
  \\fill[colorM1,cm={-1,0,0,1,(0,0)}] \\logopath;
  \\fill[colorM1,cm={-1,0,0,-1,(0,0)}] \\logopath;

  % Now place letters
  \\node[colorM5] at (-0.78,\\logotextjitter) {F};
  \\node[colorM5] at (-0.32,-\\logotextjitter) {A};
  \\node[colorM5] at (0.05,\\logotextjitter) {V};
  \\node[colorM5] at (0.4,-\\logotextjitter) {\\LARGE i};
  \\node[colorB3] at (0.79,\\logotextjitter) {R};
  \\fill[colorB3] (0.383,0.125) circle (0.06); % dot the i
\\end{tikzpicture}
\\end{minipage}}

%%%%%%%%%%%%%%%%% End Logo, Start Headers and page definition

\\usepackage{fancyhdr}
\\pagestyle{fancy}
\\renewcommand{\\headrule}{{\\color{colorM5}%
  \\hrule width\\headwidth height\\headrulewidth \\vskip-\\headrulewidth}}
\\fancyfoot[LE]{\\vspace{0ex}\\favirlogo}
\\fancyfoot[RO]{\\vspace{0ex}\\favirlogo\\hspace{2cm}}
\\fancyfoot[CE,CO]{\\vspace{2.5ex}\\thepage}

%%%%%%%%%%%%%%%% End Headers, Preamble")
}
